// Author : Nicolas JEAN
// Date   : October 2001
// filename : common.c
// Copyright : Nicolas JEAN 2001 - 2004
// available on http://youpee.salemioche.com

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "common.h"
#include "youpee.h"

char *replace_string( char *line, char *pattern, char *change) {
char buf[MAX_LINE_LENGTH];
char *match, *dest = buf;
char *zline = line;
unsigned int did_one = 0;

    while ( ( match = strstr(zline, pattern) ) )  {
      *match = 0x00;
      strcpy(dest,zline); dest += strlen(zline);
      match += strlen(pattern);
      if ( change ) {
        strcat(dest, change);
        dest += strlen( change);
      }
      zline = match;
      did_one++;
    }
    if ( did_one ) {
      strcat(dest, zline); 
      strcpy(line, buf);
      return line;
    }
    return NULL;
}

char *unescape(char *line) {
  replace_string(line, "\\\\", "$*");
  replace_string(line, "\\", "");
  replace_string(line, "$*", "\\");
  return line;
}

char *findPath(char *filename, char *includeList)
{
char *pathList;
char *lang = getenv(ENV_LANG);
char *ptr, *inc, *file;
FILE *ftest;
int eoi = FALSE;

  if ( !filename ) return NULL;
  if ( filename[0] == '!' ) return strdup(&filename[1]);
  if ( ! includeList ) pathList = strdup(".");
  else pathList = strdup(includeList);
  if ( lang == NULL ) lang = "";
  ptr = inc = pathList;
  while ( ! eoi ) {
    while ( *ptr != ':' && *ptr != 0x00) ptr++;
    if ( ! *ptr ) eoi = TRUE;
    *ptr = 0x00;
    file = malloc((strlen(lang)+strlen(inc)+1+strlen(filename)+1)*sizeof(char));
    if ( ! file ) {
      fprintf(stderr, __FILE__ ".%d: out of memory",__LINE__);
      exit(1);
    }
    sprintf(file,"%s/%s%s",inc,lang,filename);
    if ( ( ftest = fopen(file,"r") ) ) {
      fclose(ftest); free(pathList);
      return file;
    }
    sprintf(file,"%s/%s",inc,filename);
    if ( ( ftest = fopen(file,"r") ) ) {
      fclose(ftest); free(pathList);
      return file;
    }
    inc = ++ptr;
    free(file);
  }
  return strdup(filename);
}

char *get_fromfield(char *line, unsigned int number) {
char *delim;
char *field, *one_field, *cline;
char delimiter = SEPARATOR;

    delim = field = cline = strdup(line);
    do {
      if ( *field == QUOTE ) {
        delimiter = QUOTE;
        field++;
      } else
        delimiter = SEPARATOR;
      one_field = field;
      do {
        delim = strchr(one_field , delimiter);
        if ( ! delim ) break;
        one_field = delim+1;
      } while ( delim && *(delim-1) == '\\' );
      if ( delimiter == QUOTE ) {
        delimiter = SEPARATOR;
        if ( delim ) delim++;
      }
      while ( delim && *(delim+1) == SEPARATOR ) delim++;
      --number;
      if ( number && delim) field = delim + 1;
     } while ( number && delim ); 

    if ( number ) { free(cline); return ""; }// not found
    one_field = unescape(strdup(field));
    free(cline);
    return one_field;
}

char *get_fieldpos(char *line, unsigned int number) {
char *delim;
char *field, *one_field;
char delimiter = SEPARATOR;

    delim = field = line;
    do {
      if ( *field == QUOTE ) {
        delimiter = QUOTE;
        field++;
      } else
        delimiter = SEPARATOR;
      one_field = field;
      do {
        delim = strchr(one_field , delimiter);
        if ( ! delim ) break;
        one_field = delim+1;
      } while ( delim && *(delim-1) == '\\' );
      if ( delimiter == QUOTE ) {
        delimiter = SEPARATOR;
        if ( delim ) delim++;
      }
      while ( delim && *(delim+1) == SEPARATOR ) delim++;
      --number;
      if ( number && delim) field = delim + 1;
     } while ( number && delim ); 

    if ( number ) { return NULL; }// not found
    return field;
}


char *get_field(const char *line, unsigned int number) {
char *delim;
char *field, *one_field, *cline;
char delimiter = SEPARATOR;

    field = cline = strdup(line);
    delim = field;
    do {
      if ( *field == QUOTE ) {
        delimiter = QUOTE;
        field++;
      } else
        delimiter = SEPARATOR;
      one_field = field;
      do {
        delim = strchr(one_field , delimiter);
        if ( ! delim ) break;
        one_field = delim+1;
      } while ( delim && *(delim-1) == '\\' );
      if ( delim ) *delim = 0x00;
      if ( delimiter == QUOTE ) {
        delimiter = SEPARATOR;
        if ( delim ) delim++;
      }
      while ( delim && *(delim+1) == SEPARATOR ) delim++;
      --number;
      if ( number && delim) field = delim + 1;
    } while ( number && delim ); 

    if ( number )  { free(cline); return NULL; }// not found
    else {
      delim = field + strlen(field) ;
      while ( *--delim == '\r' || *delim == '\n' );
      *++delim = 0x00;
    }
    one_field = unescape(strdup(field));
    free(cline);
    return one_field;
}

char *strlower(char *s){
char *s1 = strdup(s);
char *p = s1;

  while ( ( *p++ = tolower(*p) ) );
  return s1;
}

char *removeRN(char *str) {
char *s = strdup(str);
    if ( s[strlen(s)-1] == '\n'  ||
         s[strlen(s)-1] == '\r' ) s[strlen(s)-1] = 0x00; // remove the \r or \n
    if ( s[strlen(s)-1] == '\n'  ||
         s[strlen(s)-1] == '\r' ) s[strlen(s)-1] = 0x00; // idem
    return s;
}

char getlastchar(char *str) {
char *p = str + strlen(str);
  do {
    p--;
  } while ( ( *p == '\r' || *p == '\n' ) && p > str );
  if ( p > str ) return *p;
  return 0x00;
}

char *getstrline(char *b, unsigned int l, FILE  *in) {
char *p = b, *rc;
unsigned int rl = l, ll;

  while ( (rc = fgets(p, rl, in) ) != NULL ) {
      if ( getlastchar(p) == '\\' ) {
        rl -= (ll = strlen(removeRN(p))-1);
        p += ll;
        *p = 0x00;
      } else
        break;
  }

  if ( !rc && p == b ) return NULL;
  return b;
}

int strncmp_sol(char *line, const char *sol) {
char *p = line, *f;
  
  while ( ( *p == '\t' || *p == ' ') && *p ) p++;
  f = get_field(p,1);
  if ( strlen(f) == strlen(sol) && ! strncmp(f, sol, strlen(sol)) ) {
    strcpy(line, strdup(p));
    return 0;
  }      

  return 1;
}

char *find_filepath(char *original_filename) {
char *s = strdup(original_filename);
char *p = s + strlen(s);

   while ( *(--p) != '/' ) if ( p == s ) break ;
  if ( p != s ) { *p = 0x00; p++; }
  else s = "";

  return s;
}


