// Author : Nicolas JEAN
// Date   : October 2001
// filename : youpee.c
// Copyright : Nicolas JEAN 2001 - 2004
// available on http://youpee.salemioche.com

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "youpee.h"
#include "include.h"
#include "define.h"
#include "if.h"
#include "script.h"
#include "macro.h"
#include "loop.h"
#include "post.h"
#include "special.h"
#include "file.h"
#include "common.h"
#include "date.h"
#include "arithm.h"
#include "image.h"
#include "regexp.h"

unsigned char show_dependency = FALSE;
unsigned char shell_mode = FALSE;

void create_builtin_define () ;

// the process is builtin define, first_match_mol, match_sol, match_mol/define/regexp

// match operator single on a line
matcher match_sol[] = { { DIESE_INCLUDE      , do_include          } ,
                        { DIESE_DEFINE       , add_define          } ,
                        { DIESE_DBGDEFINE    , list_replace        } ,
                        { DIESE_BINDEFINE    , add_bindefine       } ,
                        { DIESE_INCLUDEDEFINE, do_includedefine    } ,
                        { DIESE_REGEXP       , add_regexp          } ,
                        { DIESE_VAR          , add_var             } ,
                        { DIESE_LIST         , add_list            } ,
                        { DIESE_TAB          , add_tab             } ,
                        { DIESE_LET          , do_let              } ,
                        { DIESE_UNDEFINE     , del_define          } ,
                        { DIESE_DISHTMLENC   , disable_html_encode } ,
                        { DIESE_ENHTMLENC    , enable_html_encode  } ,
                        { DIESE_IF           , do_if               } ,
                        { DIESE_ELSE         , do_else             } ,
                        { DIESE_ENDIF        , do_endif            } ,
                        { DIESE_SCRIPT       , do_script           } ,
                        { DIESE_PROCESS      , do_process          } ,
                        { DIESE_CALL         , do_call             } ,
                        { DIESE_POST         , do_post             } ,
                        { DIESE_LOOP         , do_loop             } ,
                        { DIESE_MACRO        , do_createMacro      } ,
                        { DIESE_CALLMACRO    , do_callMacro        } ,
                        { DIESE_PRE          , do_pre              } ,
                        { DIESE_EXIT         , do_exit             } ,
                        { DIESE_ERROR        , do_error            } ,
                        { DIESE_ESCAPE       , do_escape           } ,
                        { DIESE_UNESCAPE     , do_unescape         } ,
                        { NULL               , NULL } };

// match operator in the middle of a line
#define FIRST_MATCH_MOL  { DIESE_FILESIZE , DIESE_ENDMOL  , do_filesize } ,\
                         { DIESE_PLUS     , DIESE_ENDMOL  , do_plus     } ,\
                         { DIESE_MOINS    , DIESE_ENDMOL  , do_moins    } ,\
                         { DIESE_FOIS     , DIESE_ENDMOL  , do_fois     } ,\
                         { DIESE_DIVIS    , DIESE_ENDMOL  , do_divis    } ,\
                         { NULL           , NULL          , NULL        }

smatcher first_match_mol[] = { FIRST_MATCH_MOL };

// match operator in the middle of a line
smatcher match_mol[] = { { DIESE_DATE     , DIESE_ENDMOL  , do_date     } ,
                         { DIESE_IMG      , DIESE_ENDMOL  , do_img      } ,
                         { DIESE_UNICODE  , DIESE_ENDMOL  , do_unicode  } ,
                         FIRST_MATCH_MOL };


char **includes = NULL;
int inc_nb = 0;
int inc_lg = 0;

char **pincludes = NULL;
int pinc_nb = 0;
int pinc_lg = 0;

char *original_filename = NULL;
char *target_filename = NULL;
char *template_filename = NULL;

void usage ( void ) {
  printf( "Usage: youpee -F | [OPTION] [-f] <original file>\n"
          "pre-process a file \n\n"
          "  -?            \t this help\n"
          "  -F            \t shell mode\n"
          "  -r            \t show file dependency on stderr\n"
          "  -t template   \t template file\n"
          "  -I include    \t include files before the org. file ( may be multiple )\n"
          "  -P include    \t include files after the org. file  ( may be multiple )\n"
          "  -D \"in out\" \t create a define  ( may be multiple )\n"
          "  -o output     \t destination file ( stdout is default )\n");
  exit(1);
}  

void add_include(char *file, char ***includes, int *nb, int *lg) {
  if ( *nb >= *lg )
    *includes = (char **) realloc(*includes, (*lg += 10) *sizeof(char *));
  (*includes)[(*nb)] = file;
  (*includes)[++(*nb)] = NULL;
}

int main (int argc, char **argv ) {
int c;
FILE *in, *out;
char **one_include;
char *env_include = getenv("INCLUDEPATH");

  while ( ( c = getopt(argc,argv,"?hrvF:f:d:o:t:I:P:D:") ) != -1 ) {
    switch ( c ) {
      case 'v' : printf("Youhp3 version %s\nhttp://youpee.salemioche.com\n", VERSION);exit(0);   break;
      case 'F' : shell_mode = TRUE;
      case 'f' : original_filename = strdup(optarg);   break;
      case 'd' : target_filename   = strdup(optarg);   break; // for backward compatibility, to be removed in 2.0
      case 'o' : target_filename   = strdup(optarg);   break;
      case 'D' : add_builtin_define(get_field(optarg,1),get_field(optarg,2)); break;
      case 'r' : show_dependency=TRUE; break;
      case 't' : template_filename = strdup(optarg);   break;
      case 'I' : add_include(findPath(strdup(optarg), env_include), &includes, &inc_nb, &inc_lg); break;
      case 'P' : add_include(findPath(strdup(optarg), env_include), &pincludes, &pinc_nb, &pinc_lg); break;
      case 'h' :
      case '?' :
      default  :
        usage();
        return 0;
    }
  }

  if ( ! original_filename ) {
    if ( optind < argc ) original_filename = argv[optind];
    else usage();
  }

  if ( ! template_filename ) template_filename = original_filename;

  if ( target_filename ) {
    out = fopen(target_filename,"w");
    if ( ! out ) {
      fprintf(stderr, __FILE__ ".%d: can't open file %s\n",__LINE__,
              target_filename);
      return 1;
    }
  } else {
    out = stdout;
    target_filename = "(stdout)";
  }
  
  create_builtin_define();
  add_include(findPath(template_filename, env_include), &includes, &inc_nb, &inc_lg);

  if ( pincludes ) {
    one_include = pincludes;
    while ( *one_include ) {
      add_include(*one_include++, &includes, &inc_nb, &inc_lg);
    }
  }
  
  one_include = includes;
  while ( *one_include ) {
    in = fopen(*one_include,"r");
    if ( ! in ) {
      fprintf(stderr, __FILE__ ".%d: can't open file %s\n",__LINE__,
              *one_include);
    } else {
      SHOW_DEP(*one_include);
      replace_builtin_define(CUR_PATH,find_filepath(*one_include));
      process_file(in, out, NULL);
      fclose(in);
    }
    one_include++;
  }

  return 0;
}

void process_line_eval_gen(char *buf, smatcher match_mol[], int with_def) {
char tempbuf[MAX_LINE_LENGTH];
smatcher *sm;
int did_one;
int level;
    do {
      sm = match_mol;
      did_one = 0;
      while ( sm->beginpattern != NULL ) {
        char *field1, *field2;
        if ( ( field1 = strstr(buf, sm->beginpattern) ) ) {
          field2=field1+strlen(sm->beginpattern)-1;
          level = 1;
          while ( ++field2 ) {
            if ( !strncmp(field2,DIESE_MOL,strlen(DIESE_MOL)) ) {
              level++;
              continue;
            }
            if ( !strncmp(field2,DIESE_ENDMOL,strlen(DIESE_ENDMOL)) ) level--;
            if ( ! level ) {
              did_one++; *field2 = 0x00;
              strncpy(tempbuf, buf, field1-buf); tempbuf[field1-buf] = 0x00;
              strcat(tempbuf, (*(sm->f))(field1) );
              strcat(tempbuf, field2 + strlen(sm->endpattern) );
              strcpy(buf, tempbuf);
              break;
            } 
          }
        }
        sm++;
      }
      if ( with_def) {
        if (do_defines(buf)) did_one++;
        if (do_builtin_defines(buf)) did_one++;
        do_regexp(buf);
      }
    } while ( did_one );
}

void process_line_eval(char *buf) {
  process_line_eval_gen(buf, match_mol, TRUE);
}

int process_line(char *buf,FILE *in, FILE *out) {
char *str;
matcher *m;

    do_builtin_defines(buf);
    process_line_eval_gen(buf, first_match_mol, FALSE);
    str = buf; while ( *str == SEPARATOR ) str++;
    m = match_sol;
    while ( m->pattern != NULL ) {
      if ( ! strncmp_sol(str, m->pattern) ) {
        fflush(out);
        (*(m->f))(in,out,str);
        break;
      }
      m++;
    }
    if ( m->pattern != NULL) return FALSE;
    process_line_eval(buf);
    return TRUE;
}

void process_file(FILE *in, FILE *out, char *endtag) {
char buf[MAX_LINE_LENGTH];
unsigned int le;

  if (endtag) le = strlen(endtag);

  while ( getstrline(buf, MAX_LINE_LENGTH, in) != NULL ) 
    if ( ! strncmp(buf,BREAK,strlen(BREAK))  ||
         ( endtag && !strncmp(buf,endtag,le) ) ) return;
    else if ( shell_mode && buf[0] == '#' && buf[1] == '!' ) continue;
    else if ( process_line(buf, in, out) ) fputs(buf, out);
  
  fflush(out);
}

char *findRootFilename(char *fn) {
char *c = fn+strlen(fn)-1;
  while ( *c != '.' && c >= fn && *c != '/' ) c--;
  if ( c >= fn && *c != '/' ) *c = 0x00;
  return fn;
}

void create_builtin_define (void) 
{
char *s = strdup(original_filename);
char *p = s + strlen(s);
char *t = findRootFilename(strdup(target_filename));
char *tn = (char *) malloc(sizeof(char)*(strlen(t)+2)); // if 9 then 10 (so 1 more char) + 0x00
char *p1 = t + strlen(t);
int n=0;

  while ( *(--p) != '/' ) if ( p == s ) break ;
  if ( p != s ) { *p = 0x00; p++; }
  else s = "";
  
  while ( *(--p1) != '/' ) if ( p1 == t ) break ;
  if ( p1 != t ) p1++;
  t = p1;
  p1 = t + strlen(t);

  while ( p1 > t ) {
    p1--;
    if ( *p1 < '0' || *p1 > '9' || p1 == t) {
      n = atoi(p1+1)+1;
      *(p1+1) = 0x00;
      break;
    }
  }
  
  sprintf(tn,"%s%d",t,n);

  add_builtin_define(YOUPEE_NAME, YOUPEE_NAME_VALUE);
  add_builtin_define(ORG_SHORTFILENAME,p);
  add_builtin_define(ORG_PATH,s);
  add_builtin_define(CUR_PATH,s);
  add_builtin_define(ORG_FILENAME, original_filename);
  add_builtin_define(TARGET_FILENAME, target_filename);
  add_builtin_define(TEMPLATE_FILENAME, template_filename);
  add_builtin_define(ORG_ROOTFILENAME, findRootFilename(strdup(original_filename)));
  add_builtin_define(TARGET_ROOTFILENAME, findRootFilename(strdup(target_filename)));
  add_builtin_define(NTARGET_ROOTFILENAME, tn);
  add_builtin_define(TEMPLATE_ROOTFILENAME, findRootFilename(strdup(template_filename)));
  add_builtin_define(LANG, getenv(ENV_LANG));
  enable_html_encode(NULL, NULL, NULL);
}
